/*
glass - (web)gl-assistant.
Copyright (C) 2011  John Tsiombikas <nuclear@member.fsf.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* public interface

bool glass_init(canvas-id, optional-console-textarea-id);
void glass_start(void);

void glass_display_func(callback);
void glass_reshape_func(callback);
void glass_mouse_func(callback);
void glass_motion_func(callback);

void glass_resize(x, y);

WebGLTexture load_texture(uri);

WebGLShader load_shader(uri, type);
WebGLShader create_shader(source, type, optional-sdrname);
WebGLProgram load_program(vertex-uri, fragment-uri);
WebGLProgram create_program(vertex-shader, fragment-shader);

void set_uniform1i(program, uniform-name, value);
void set_uniform1f(program, uniform-name, value);
void set_uniform2f(program, uniform-name, x, y);
void set_uniform3f(program, uniform-name, x, y, z);
void set_uniform4f(program, uniform-name, x, y, z, w);
void set_uniform_matrix4fv(program, uniform-name, matrix);

void request_redisplay(void);

void logmsg(string);
*/

var gl;
var canvas, console;
var glass_orig_width, glass_orig_height;
var glass_data_loaded, glass_data_count;
var glass_progr_prog;
var glass_start_time = 0;
var glass_fullscr = false;

var glass_cb_disp = null;
var glass_cb_reshape = null;
var glass_cb_mouse = null;
var glass_cb_motion = null;

function glass_init(canv_id, cons_id)
{
	if(cons_id && (console = document.getElementById(cons_id))) {
		console.value = "";
	}

	if(!(canvas = document.getElementById(canv_id))) {
		logmsg("canvas \"" + canv_id + "\" does not exist\n");
		return false;
	}
	try {
		gl = canvas.getContext("experimental-webgl");
	}
	catch(e) {
		logmsg("can't get an OpenGL context\n");
		return false;
	}

	glass_orig_width = canvas.width;
	glass_orig_height = canvas.height;

	var vs = create_shader(glass_progr_vs_src, gl.VERTEX_SHADER);
	var ps = create_shader(glass_progr_ps_src, gl.FRAGMENT_SHADER);
	if(!vs || !ps || !(glass_progr_prog = create_program(vs, ps))) {
		logmsg("internal glass error\n");
		return false;
	}

	glass_data_loaded = glass_data_count = 0;
	return true;
}

function glass_start()
{
	if(glass_cb_motion) {
		canvas.onmousemove = function(ev) {
			glass_cb_motion(ev.clientX, ev.clientY);
		}
	}

	if(glass_cb_mouse) {
		canvas.onmousedown = function(ev) {
			glass_cb_mouse(ev.button, true, ev.clientX, ev.clientY);
		}
		canvas.onmouseup = function(ev) {
			glass_cb_mouse(ev.button, false, ev.clientX, ev.clientY);
		}
	}

	glass_start_time = get_msec();

	if(glass_cb_reshape) {
		glass_cb_reshape(canvas.width, canvas.height);
	}
	glass_redraw();
}

function glass_display_func(func)
{
	glass_cb_disp = func;
}

function glass_reshape_func(func)
{
	glass_cb_reshape = func;
}

function glass_mouse_func(func)
{
	glass_cb_mouse = func;
}

function glass_motion_func(func)
{
	glass_cb_motion = func;
}

function glass_resize(x, y)
{
	canvas.width = glass_orig_width = x;
	canvas.height = glass_orig_height = y;

	if(glass_cb_reshape) {
		glass_cb_reshape(x, y);
	}
}

function glass_fullscreen(state)
{
	if(state == null) {
		glass_fullscr = !glass_fullscr;
	} else {
		glass_fullscr = state;
	}

	if(glass_fullscr) {
		canvas.style.position = "absolute";
		canvas.style.left = 0;
		canvas.style.top = 0;
		canvas.width = window.innerWidth;
		canvas.height = window.innerHeight;

		window.onresize = function() {
			canvas.width = window.innerWidth;
			canvas.height = window.innerHeight;
			if(glass_cb_reshape) {
				glass_cb_reshape(canvas.width, canvas.height);
			}
		}
	} else {
		canvas.style.position = "static";
		canvas.width = glass_orig_width;
		canvas.height = glass_orig_height;

		window.onresize = null;
	}

	if(glass_cb_reshape) {
		glass_cb_reshape(canvas.width, canvas.height);
	}
}


function glass_redraw()
{
	var msec = get_msec();

	if(glass_data_loaded < glass_data_count || !glass_cb_disp) {
		glass_draw_progress(msec);
	} else {
		glass_cb_disp(msec);
	}
	request_redisplay(glass_redraw);
}

function get_msec()
{
	return new Date().getTime() - glass_start_time;
}

/* ---- textures ---- */
function load_texture(name)
{
	glass_data_count++;

	var tex = gl.createTexture();
	tex.image = new Image();
	tex.image.onload = function() { glass_tex_load_done(tex); }
	tex.image.onerror = glass_tex_load_failed;
	tex.image.src = name;
	return tex;
}

function glass_tex_load_done(tex)
{
	gl.bindTexture(gl.TEXTURE_2D, tex);
	gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, tex.image);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.REPEAT);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.REPEAT);
	gl.bindTexture(gl.TEXTURE_2D, null);
	glass_data_loaded++;
	logmsg("loaded image: " + tex.image.src +
			" (" + glass_data_loaded + "/" + glass_data_count + ")\n");
}

function glass_tex_load_failed()
{
	logmsg("failed to load texture: " + this.src + "\n");
}


/* ---- shaders ---- */
function load_shader(name, type)
{
	var xhr = new XMLHttpRequest();
	xhr.open("GET", name, false);
	xhr.overrideMimeType("text/plain");
	xhr.send(null);

	var src = xhr.responseText;
	return create_shader(src, type, name);
}

function create_shader(src, type, name)
{
	var sdr = gl.createShader(type);
	gl.shaderSource(sdr, src);
	gl.compileShader(sdr);

	if(!gl.getShaderParameter(sdr, gl.COMPILE_STATUS)) {
		if(!name) {
			name = "<unnamed>";
		}
		logmsg("failed to compile shader: " + name + ": " + gl.getShaderInfoLog(sdr));
		return null;
	}
	return sdr;
}

function load_program(vsname, psname)
{
	var vs, ps;

	if(!(vs = load_shader(vsname, gl.VERTEX_SHADER))) {
		return null;
	}
	if(!(ps = load_shader(psname, gl.FRAGMENT_SHADER))) {
		return null;
	}
	return create_program(vs, ps);
}

function create_program(vs, ps)
{
	var prog = gl.createProgram();
	gl.attachShader(prog, vs);
	gl.attachShader(prog, ps);
	gl.linkProgram(prog);

	if(!gl.getProgramParameter(prog, gl.LINK_STATUS)) {
		logmsg("Failed to link GLSL program (" + vsname + " - " + psname + ")\n");
		return null;
	}
	return prog;
}

function set_uniform1i(p, name, v)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniform1i(loc, v);
	}
}

function set_uniform1f(p, name, v)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniform1f(loc, v);
	}
}

function set_uniform2f(p, name, x, y)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniform2f(loc, x, y);
	}
}

function set_uniform3f(p, name, x, y, z)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniform3f(loc, x, y, z);
	}
}

function set_uniform4f(p, name, x, y, z, w)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniform4f(loc, x, y, z, w);
	}
}

function set_uniform_matrix4fv(p, name, mat)
{
	var loc = gl.getUniformLocation(p, name);
	if(loc != -1) {
		gl.useProgram(p);
		gl.uniformMatrix4fv(loc, false, mat);
	}
}

function logmsg(str)
{
	if(console) {
		console.value += str;
	}
}

window.request_redisplay = (function() {
	return	window.requestAnimationFrame ||
			window.webkitRequestAnimationFrame ||
			window.mozRequestAnimationFrame ||
			window.oRequestAnimationFrame ||
			window.msRequestAnimationFrame ||
			function(callback, element) { window.setTimeout(callback, 1000/60); };
})();


var glass_progr_vs_src = 
	"attribute vec4 attr_vertex, attr_color;\n" +
	"uniform mat4 mvmat, projmat;\n" +
	"varying vec4 color;\n" +
	"void main()\n{\n" +
	"\tmat4 mvp = projmat * mvmat;\n" +
	"\tgl_Position = mvp * attr_vertex;\n" +
	"\tcolor = attr_color;\n}\n";

var glass_progr_ps_src =
	"precision highp float;\n" +
	"varying vec4 color;\n" +
	"void main()\n{\n" +
	"\tgl_FragColor = color;\n}\n";

function glass_draw_progress(msec)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	gl.clearColor(0, 0, 0, 1);
	gl.clear(gl.COLOR_BUFFER_BIT);

	var progr = glass_data_loaded / glass_data_count;

	gl.useProgram(glass_progr_prog);

	glBegin(GL_QUADS);
	glColor3f(0.3, 0.5, 0.8);
	glVertex2f(-0.55, -0.1);
	glVertex2f(0.55, -0.1);
	glVertex2f(0.55, 0.1);
	glVertex2f(-0.55, 0.1);

	glColor3f(0, 0, 0);
	glVertex2f(-0.525, -0.075);
	glVertex2f(0.525, -0.075);
	glVertex2f(0.525, 0.075);
	glVertex2f(-0.525, 0.075);

	glColor3f(1.0, 0.3, 0.2);
	glVertex2f(-0.5, -0.05);
	glVertex2f(progr - 0.5, -0.05);
	glVertex2f(progr - 0.5, 0.05);
	glVertex2f(-0.5, 0.05);
	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

